#include <opencv2/opencv.hpp>

namespace {

    void GaussianBlurViaFFT(const cv::Mat &src, cv::Mat &dst, const double sigma)
    {
        int ksize = cvRound(sigma*3*2 + 1)|1; // sigma to radius
        cv::Mat kernel1D = cv::getGaussianKernel(ksize,sigma,CV_32F),kernel1D_T;
        cv::transpose(kernel1D,kernel1D_T);
        cv::Mat kernel2D = kernel1D * kernel1D_T;
        cv::Mat _src, _dst;
        src.convertTo(_src,CV_32F);
        cv::filter2D(_src,_dst,CV_32F,kernel2D,cv::Point(-1,-1),0.0,cv::BORDER_REPLICATE);
        _dst.convertTo(dst,src.depth());
    }

    void singleScaleRetinex(const cv::Mat &src, cv::Mat &dst, double sigma)
    {
        std::vector<cv::Mat> src_;
        cv::split(src, src_);
        for (int i = 0; i < src.channels(); i++) {
            cv::Mat src_blured, log_src, log_src_blured;
            GaussianBlurViaFFT(src_[i], src_blured, sigma);
            cv::log(src_[i], log_src);
            log_src /= log(10.0);
            cv::log(src_blured, log_src_blured);
            log_src_blured /= log(10);
            cv::subtract(log_src, log_src_blured, src_[i]);
        }
        cv::merge(src_, dst);
    }

    void multiScaleRetinex(const cv::Mat &src, cv::Mat &dst, std::vector<double> &sigma_list)
    {
        cv::Mat retinex = cv::Mat(src.size(), CV_64FC3, cv::Scalar(0.0));
        for (double sigma : sigma_list) {
            cv::Mat single;
            singleScaleRetinex(src, single, sigma);
            retinex += single;
        }
        retinex /= float(sigma_list.size());
        retinex.copyTo(dst);
    }

    void colorRestoration(const cv::Mat &img, cv::Mat &color_restoration, double alpha, double beta)
    {
        std::vector<cv::Mat> img_;
        cv::split(img, img_);
        cv::Mat img_sum(img.size(), CV_64FC1, cv::Scalar(0.0));
        for (int i = 0; i < img.channels(); i++) {
            img_[i].convertTo(img_[i], CV_64F);
            img_sum += img_[i];
            cv::log(alpha*img_[i], img_[i]);
            img_[i] /= cv::log(10.0);
        }
        cv::log(img_sum, img_sum);
        img_sum /= cv::log(10.0);
        for (int i = 0; i < img.channels(); i++) {
            img_[i] = beta * (img_[i] - img_sum);
            img_[i].convertTo(img_[i], CV_8U);
        }
    
        cv::merge(img_, color_restoration);
    }

    void simplestColorBalance(const cv::Mat &img, cv::Mat &balanced, double low_clip, double high_clip)
    {
        int total = img.rows*img.cols;
        std::vector<cv::Mat> img_;
        cv::split(img, img_);
        for (int i = 0; i < img.channels(); i++) {
            std::vector<int> hist(256, 0);
            for (cv::MatIterator_<uchar> it = img_[i].begin<uchar>(); it != img_[i].end<uchar>(); it++) hist[*it]++;
            int current = 0;
            int low_val = 0, high_val = 255;
            for (int u = 0; u < 256; u++) {
                if (hist[u] == 0) continue;
                int c = hist[u];
                if (float(current) / total < low_clip) low_val = u;
                if (float(current) / total < high_clip) high_val = u;
                current += c;
            }
            img_[i].setTo(cv::Scalar(low_val), img_[i] < low_val);
            img_[i].setTo(cv::Scalar(high_val), img_[i] > high_val);
        }
        cv::merge(img_, balanced);
    }

    void MSRCR(const cv::Mat &src, cv::Mat &dst, std::vector<double> &sigma_list, double G, double b, double alpha, double beta, double low_clip, double high_clip)
    {
        cv::Mat img;
        src.convertTo(img, CV_64F);
        img += 1.0;

        cv::Mat img_retinex;
        multiScaleRetinex(img, img_retinex, sigma_list);

        cv::Mat img_color;
        colorRestoration(img, img_color, alpha, beta);
        img_color.convertTo(img_color, CV_64F);

        cv::Mat img_msrcr;
        cv::multiply(img_retinex, img_color, img_msrcr);
        cv::add(img_msrcr,b, img_msrcr);
        cv::multiply(img_msrcr, G, img_msrcr);

        std::vector<cv::Mat> img_msrcr_(img_msrcr.channels());
        cv::split(img_msrcr, img_msrcr_);
        for (int i = 0; i < img_msrcr.channels(); i++) {
            double minimum = std::numeric_limits<double>::max();
            double maximum = std::numeric_limits<double>::min();
            for (cv::MatIterator_<double> it = img_msrcr_[i].begin<double>(); it != img_msrcr_[i].end<double>(); it++) {
                if (*it < minimum) minimum = *it;
                if (*it > maximum) maximum = *it;
            }
            img_msrcr_[i] = 255 * (img_msrcr_[i] - minimum) / (maximum - minimum);
            img_msrcr_[i].setTo(cv::Scalar(0.0), img_msrcr_[i] < 0.0);
            img_msrcr_[i].setTo(cv::Scalar(255.0), img_msrcr_[i] > 255.0);
        }
        cv::merge(img_msrcr_, img_msrcr);
        img_msrcr.convertTo(img_msrcr, CV_8U);
        simplestColorBalance(img_msrcr, dst, low_clip, high_clip);
    }

}

int main1() {
    cv::Mat img = cv::imread("../DinningRoom.jpg", cv::IMREAD_COLOR), img_msrcr;

    std::vector<double> sigma_list{ 15, 80, 250 };
    double G = 5.0;
    double b = 25.0;
    double alpha = 125.0;
    double beta = 46.0;
    double low_clip = 0.01;
    double high_clip = 0.99;
    MSRCR(img, img_msrcr, sigma_list, G, b, alpha, beta, low_clip, high_clip);

    cv::imwrite("intrinsic-DinningRoom.jpg", img_msrcr);
    return 0;
}